# ===============================================================
# TestMultiLayerNeuralNetwork01.py: Train and test neural network
# with one hidden layer for data that is non-separable.
# 
# ----------------------------------
#    Input 1     Input 2      Output 
# ----------------------------------
#          0           0           0 
#          0           1           1        
#          1           0           1        
#          1           1           0        
# ----------------------------------
# 
# We cannot simply classify the data with a straight line ... 
# 
# --------------------------------------------------------------- 
# Network Architecture:
# --------------------------------------------------------------- 
# 
# 2 input nodes, 4 nodes on hidden layer, 1 node on output. 
# 
# Weighted edges: 
# 
# input  layer --> hidden layer: connected by 2x4 = 8 weights. 
# hidden layer --> output layer: connected by 4 weights. 
# 
# Reference: Shukla, et al., Neural Networks from Scratch with
# Python Code and Math in Detail, Towards AI, 2020  
# 
# Modified by: Mark Austin                          October, 2020 
# ===============================================================

import math
import matplotlib
import matplotlib.pyplot as plt
import numpy as np

# Define Sigmoid function:

def sigmoid(x):
   return 1/(1+np.exp(-x))

# Define derivative of Sigmoid function:

def sigmoid_der(x):
   return sigmoid(x)*(1-sigmoid(x))

# main method ...

def main():
    print("--- Enter TestMultiLayerNeuralNetwork01.main() ... ");
    print("--- ========================================== ... ");

    print("--- ");
    print("--- Part 1: Define input features ... ");

    input_features = np.array( [[0,0],[0,1],[1,0],[1,1]] )

    print (input_features.shape)
    print (input_features)

    # Define target output:

    print("--- ");
    print("--- Part 2: Define target output ... ");
    print("--- ");

    target_output = np.array([[0,1,1,0]])

    # Reshaping target output into vector:

    target_output = target_output.reshape(4,1)
    print (target_output)

    print("--- ");
    print("--- Part 3: Initial weights and learning parameters ... ");
    print("--- ");
    
    weight_hidden = np.random.rand(2,4)
    weight_output = np.random.rand(4,1)

    print(weight_hidden)
    print(weight_output)

    lr   = 0.05 # Learning Rate:

    # ========================================================
    # Main logic for neural network:
    # ========================================================

    print("--- ");
    print("--- Part 4: Train Neural Network ... ");
    print("--- ---------------------------- ... ");
    print("--- ");

    # Running our code 300000 times:

    for epoch in range(300000):

       print("--- Start epoch: %f ..." %(epoch))

       # Input for hidden layer :

       print("---   Compute input for hidden layer ...")

       input_hidden = np.dot(input_features, weight_hidden)
 
       # Output from hidden layer :

       print("---   Compute output from hidden layer ...")

       output_hidden = sigmoid(input_hidden)
 
       # Input for output layer :

       print("---   Compute input for output layer ...")

       input_op = np.dot(output_hidden, weight_output)
 
       # Output from output layer :

       print("---   Compute output for output layer ...")

       output_op = sigmoid(input_op)

       # ======================================================
       # Phase 1: Compute derivatives for phase 1 ...
       # ======================================================
 
       # Calculating Mean Squared Error :

       print("---   Compute mean square error ...")

       error_out = ((1 / 2) * (np.power((output_op - target_output), 2)))

       print("---   Mean square error: %f ... \n" %( error_out.sum() ) )
 
       # Derivatives for phase 1 :

       print("---   Compute derivatives for Phase 1 ...")

       derror_douto = output_op - target_output
       douto_dino   = sigmoid_der(input_op) 
       dino_dwo     = output_hidden

       derror_dwo = np.dot(dino_dwo.T, derror_douto * douto_dino)

       # ======================================================
       # Phase 2:
       # 
       # derror_w1    = derror_douth * douth_dinh * dinh_dw1
       # derror_douth = derror_dino * dino_outh
       # ======================================================
 
       # Derivatives for phase 2 :

       print("---   Compute derivatives for Phase 2 ...")

       derror_dino  = derror_douto * douto_dino
       dino_douth   = weight_output
       derror_douth = np.dot(derror_dino , dino_douth.T)

       douth_dinh = sigmoid_der(input_hidden) 
       dinh_dwh   = input_features
       derror_dwh = np.dot(dinh_dwh.T, douth_dinh * derror_douth)

       print("---   Update hidden layer and output layer weights ...")

       weight_hidden -= lr * derror_dwh
       weight_output -= lr * derror_dwo

       # Summary of iteration results ...

       # print("---   Summary of iteration results ... ");
       # print( weight_hidden ) 
       # print( weight_output ) 

    # ============================================
    # Print Results ...
    # ============================================

    print("--- ");
    print("--- ------------------------------------- ... ");
    print("--- Part 5: Summary of weights            ... ");
    print("--- ------------------------------------- ... ");
    print("--- Final weights in hidden layer:");
    print("--- ");

    print (weight_hidden) 

    print("--- ");
    print("--- Final weights in output layer:");
    print("--- ");

    print (weight_output)

    print("--- ");
    print("--- --------------------------------------------- ... ");
    print("--- Part 6: Use trained network to predict values ... ");
    print("--- --------------------------------------------- ... ");

    print("--- ");
    print("--- Part 6.1: Verify input [0,0] --> 0 ... ");
    print("--- ");

    single_point = np.array([0,0]) 

    result1 = np.dot(single_point, weight_hidden) 
    result2 = sigmoid(result1)
    result3 = np.dot( result2, weight_output)
    result4 = sigmoid(result3)

    print ("--- Single input point [0,0] ...")
    print ("--- Result 4: %f ..." %(result4))

    print("--- ");
    print("--- Part 6.2: Verify input [1,0] --> 1 ... ");
    print("--- ");

    single_point = np.array([1,0])

    result1 = np.dot(single_point, weight_hidden) 
    result2 = sigmoid(result1)
    result3 = np.dot(result2, weight_output)
    result4 = sigmoid(result3)

    print ("--- Single input point [1,0] ...")
    print ("--- Result 4: %f ..." %(result4))

    print("--- ");
    print("--- Part 6.3: Verify input [0,1] --> 1 ... ");
    print("--- ");

    single_point = np.array([0,1])

    result1 = np.dot(single_point, weight_hidden) 
    result2 = sigmoid(result1)
    result3 = np.dot(result2, weight_output)
    result4 = sigmoid(result3)

    print ("--- Single input point [0,1] ...")
    print ("--- Result 4: %f ..." %(result4))

    print("--- ");
    print("--- Part 6.4: Verify input [1,1] --> 0 ... ");
    print("--- ");

    single_point = np.array([1,1])

    result1 = np.dot(single_point, weight_hidden) 
    result2 = sigmoid(result1)
    result3 = np.dot(result2, weight_output)
    result4 = sigmoid(result3)

    print ("--- Single input point [1,1] ...")
    print ("--- Result 4: %f ..." %(result4))

    print("--- ============================================= ... ");
    print("--- Finished TestMultiLayerNeuralNetwork01.main() ... ");

# call the main method ...

main()
